#!/usr/bin/env python
# vim: fileencoding=utf-8

"""
phn_output.py (Inkscape extension)

Copyright (C) 2011 Tatt61880 (tatt61880@gmail.com, @tatt61880)
Last Change: 2011/02/13 22:08:34.

"""
#TODO 自己交差しているパスや複合パスへの対応
#TODO fill-ruleがnonzero で自己交差している場合に警告を出力
#TODO use要素に対応
#TODO Scene.Cameraに対応
#TODO Textureに対応

VERSION = '0.0.2b'

# ベジエ曲線を線分に分割する際の精度。1つのベジエ曲線あたりの点数。
NODES = 16

import inspect # for __LINE__ info
import sys, re
from math import *
from random import random
from xml.etree.ElementTree import ElementTree, _ElementInterface

xmlns = '{http://www.w3.org/2000/svg}'
sodipodi = '{http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd}'

ORIGINAL_DATA_OUTPUT = False

# MINIMUM_MODEをFalseにすると、デフォルトのパラメータも出力する。例えば density = 2.0など。
# Algodoo用のパラメータを多数出力してしまうため、MINIMUM_MODE=Trueにしておくのがオススメ。
# FIXME MINIMUM_MODEは全くテストしていません。
MINIMUM_MODE = True

def warning(description): #{{{
    c = inspect.currentframe(1)
    line_no = c.f_lineno
    filename = c.f_code.co_filename
    sys.stderr.write("\n## Warning - %s line %s: \n%s\n" % (
        __name__, str(line_no), description) )
#}}}

# ========================================
# Classes
# ========================================
class Matrix(tuple): #{{{
    def combine(self, other):
        if isinstance(other, self.__class__):
            return Matrix([
                self[0]*other[0] + self[2]*other[1],
                self[1]*other[0] + self[3]*other[1],
                self[0]*other[2] + self[2]*other[3],
                self[1]*other[2] + self[3]*other[3],
                self[0]*other[4] + self[2]*other[5] + self[4],
                self[1]*other[4] + self[3]*other[5] + self[5],
                ])
        else:
            raise ValueError('Argument should be %s class' % self.__class__.__name__)
    def __init__(self, data):
        if len(self) != 6:
            raise ValueError('%s must be 6 element array.' % self.__class__.__name__)
#}}}
class Transforms(list):#{{{
    def _transform_to_matrix(self, transform):#{{{
        '''
        単一の transformを、matrixに変換
        transform: matrix | translate | scale | rotate | skewX | skewY
        '''
        label = transform[0]
        params = transform[1]

        if(label == 'matrix'):
            if len(params)==6: matrix = params
        elif(label == 'translate'):
            if len(params)==1: params.append(0)
            if len(params)==2: matrix = (1, 0, 0, 1, params[0], params[1])
        elif(label == 'scale'):
            if len(params)==1: params.append(params[0])
            if len(params)==2: matrix = (params[0], 0, 0, params[1], 0, 0)
        elif(label == 'rotate'):
            if len(params)==1:
                theta = params[0]*pi/180.
                matrix = (cos(theta), sin(theta), -sin(theta), cos(theta), 0, 0)
            elif len(params)==3:
                # rotate(<rotate-angle> <cx> <cy>)は下記と同等
                # translate(<cx>, <cy>) rotate(<rotate-angle>) translate(-<cx>, -<cy>)
                transforms = Transforms([
                        ['translate', [params[1], params[2]]],
                        ['rotate', [params[0]]],
                        ['translate', [-params[1], -params[2]]]
                        ])
                matrix = transforms.to_matrix()
        elif(label == 'skewX'):
            if len(params)==1:
                theta = params[0]*pi/180.
                matrix = (1, 0, tan(theta), 1, 0, 0)
        elif(label == 'skewY'):
            if len(params)==1:
                theta = params[0]*pi/180.
                matrix = (1, tan(theta), 0, 1, 0, 0)
        else:
            warning("Unknown label for transform ('%s')." % label)
            matrix = (1,0,0,1,0,0)

        try:
            return Matrix(matrix)
        except:
            warning("ValueError for transform <%s%s>." % (label, tuple(params)))
            return Matrix((1,0,0,1,0,0))
#}}}

    def to_matrix(self):
        matrix = Matrix((1,0,0,1,0,0))
        for transform in self:
            matrix = matrix.combine(self._transform_to_matrix(transform))
        return Matrix(matrix)

    def __init__(self, transforms_notation):
        if(isinstance(transforms_notation, str)):
            if re.match(r'^ *(\w+ *\(.*?\)( +,? *|, *))*\w+ *\(.*?\) *$', transforms_notation):
                transforms = re.findall(r'\w+ *\(.*?\)', transforms_notation)

                ret = []
                if transforms:
                    try:
                        for transform in transforms:
                            match_ = re.search(r'(\w+) *\((.*)\)', transform)
                            type = match_.group(1)
                            try:
                                params = [float(param) for param in re.split(r' +,? *|, *', match_.group(2))]
                            except:
                                warning("Strange transform - <%s>." % transform)
                                return None
                            ret.append([type, params])
                    except:
                        warning("ValueError for transform - <%s>." % transforms_notation)
                self= list.__init__(self, ret)
            else:
                warning("ValueError for transform - <%s>." % transforms_notation)
        elif(isinstance(transforms_notation, list)):
            self= list.__init__(self, transforms_notation)
        elif transforms_notation == None:
            return None
        else:
            warning("ValueError for transform - <%s>." % transforms_notation)
            return None
#}}}
class NumTuple(tuple): #{{{
    def __repr__(self):
        ret = '['
        for item in self:
            ret += "%g" % item + ', '
        return ret[:-2] + ']'

#}}}
class Color(NumTuple): #{{{
    def __init__(self, data):
        if len(self) != 4:
            raise ValueError('Color must be 4 element array ([r,g,b,a]).')
#}}}
class Vector(NumTuple):#{{{
    def en_matrix(self, m):
        '''transformのmatrix操作'''
        return Vector([m[0]*self[0]+m[2]*self[1]+m[4],
                       m[1]*self[0]+m[3]*self[1]+m[5]])

    def rotate(self, theta):
        return Vector([cos(theta)*self[0]+sin(theta)*self[1],
                       -sin(theta)*self[0]+cos(theta)*self[1]])

    def angle_between(self, other):
        if isinstance(other, Vector):
            temp = acos(self*other/(abs(self)*abs(other)))
            if self[0]*other[1]-self[1]*other[0] < 0:  #FIXME for generalization
                temp *= -1
            return temp
        else:
            raise ValueError('Strange instance')

    def __pos__(self):
        return self
    def __neg__(self):
        return Vector([-a for a in self])

    def __add__(self, other):
        if len(self) == len(other):
            return Vector([a+b for a, b in zip(self, other)])
        else:
            raise ValueError('Vector dimension error.')
    def __sub__(self, other):
        return self.__add__(-other)
    def __mul__(self, other):
        if isinstance(other, Vector):
            if len(self) == len(other):
                return sum([a*b for a, b in zip(self, other)])
            else:
                raise ValueError('Vector dimension error.')
        else:
            try:
                return Vector([other*a for a in self])
            except:
                raise ValueError('Strange value.')
    def __div__(self, other):
        try:
            return Vector([a/other for a in self])
        except:
            raise ValueError('Strange value.')

    def __abs__(self):
        return sqrt(sum([a*a for a in self]))
#}}}

class SVG_Style(dict):#{{{
    def get_stroke_color(self):
        r,g,b = string_to_rgb(self['stroke'])
        o = float(self['stroke-opacity'])
        o *= float(self['opacity'])
        return Color([r,g,b,o])

    def get_color_and_drawborder(self):
        drawBorder = False
        fill = self['fill']
        if fill=='none':
            if self['stroke'] == 'none':
                stroke = '#000000'
            else:
                stroke = self['stroke']
                drawBorder = True
            r, g, b = string_to_rgb(stroke)
            o = 0
        else:
            r, g, b = string_to_rgb(fill)
            o = float(self['fill-opacity'])

        o *= float(self['opacity'])

        return Color([r, g, b, o]), drawBorder

    def inherit(self, inherited_style):
        if inherited_style:
            ret = inherited_style.copy()
            ret.update(self)
            if 'opacity' in inherited_style and 'opacity' in self:
                ret['opacity'] = float(inherited_style['opacity']) * float(self['opacity'])
        else:
            ret = self
        return SVG_Style(ret.copy())
    def __init__(self, arg={}):
        #attributes = {'fill':'black', 'fill-opacity':'1', 'stroke':'none', 'stroke-opacity':'1', 'opacity':'1'}
        attributes = {}
        if isinstance(arg, dict):
            attributes.update(arg)
        elif isinstance(arg, _ElementInterface):
            for attribute in ('fill', 'fill-opacity', 'stroke', 'stroke-opacity', 'opacity'):
                if arg.get(attribute):
                    attributes.update({attribute:arg.get(attribute)})
            style = arg.get('style')
            if style:
                style = re.sub(r'\s*', '', style)
                styles = {}
                for property in style.split(';'):
                    if not property: continue
                    try:
                        property_name, value = property.split(':')
                    except ValueError: #style="a"など変則的な(不適切な)style
                        warning("ValueError for style <%s>." % style)
                        styles = {}
                        break
                    styles[property_name] = value
                attributes.update(styles)
        else:
            raise ValueError('Argument\'s instance should be dict or instance')
        self.update(attributes)
#}}}
class _D_TOKEN(object): #{{{
    NUMBER, COMMAND  = xrange(2)
    __slots__ = []
#}}}
#{{{
patterns = [
    (_D_TOKEN.NUMBER, re.compile(r'[+-]?(\d*\.\d+|\d+\.|\d+)([eE][+-]?\d+)?')),
    (_D_TOKEN.COMMAND, re.compile(r'[MmZzLlHhVvCcSsQqTtAa]')),
]
whitespace = re.compile('\s+')
#}}}
class SVG_Path(list): #{{{
    ###################################################
    # http://www.w3.org/TR/SVG/paths.html#PathElement #
    ###################################################
    #horizontal_lineto:               ( "H" | "h" ) wsp* number_sequence
    #vertical_lineto:                 ( "V" | "v" ) wsp* number_sequence

    #closepath:                       ( "Z" | "z" )

    #moveto:                          ( "M" | "m" ) wsp* coordinate_sequence
    #lineto:                          ( "L" | "l" ) wsp* coordinate_sequence

    #curveto:                         ( "C" | "c" ) wsp* coordinate_3times_sequence
    #smooth_curveto:                  ( "S" | "s" ) wsp* coordinate_2times_sequence

    #quadratic_bezier_curveto:        ( "Q" | "q" ) wsp* coordinate_2times_sequence
    #smooth_quadratic_bezier_curveto: ( "T" | "t" ) wsp* coordinate_sequence

    #elliptical_arc:                  ( "A" | "a" ) wsp* elliptical_arc_argument_sequence
    def _simplify(self, list):
        '''
        Changes lower-case-command to upper-case-command
        Changes H-command to L-command
        Changes V-command to L-command
        '''
        if list == []:
            return []
        ret = []

        p_last = Vector([0, 0])
        start_point_flag = True
        while list:
            command_group = list.pop(0)
            if   command_group[0] == 'H':
                command_group = ['L', command_group[1], p_last[1]]
            elif command_group[0] == 'h':
                command_group = ['L', p_last[0]+command_group[1], p_last[1]]
            elif command_group[0] == 'V':
                command_group = ['L', p_last[0], command_group[1]]
            elif command_group[0] == 'v':
                command_group = ['L', p_last[0], p_last[1]+command_group[1]]

            if   re.match(r'[Zz]', command_group[0]):
                p_last = start_point
                start_point_flag = True
                ret.append(['Z'])
                continue

            elif re.match(r'[ML]', command_group[0]):
                p_last = Vector((command_group[1], command_group[2]))
                ret.append([command_group[0], p_last])
            elif re.match(r'[ml]', command_group[0]):
                p_last += Vector((command_group[1], command_group[2]))
                ret.append([command_group[0].upper(), p_last])

            # The cubic Bezier curve commands
            elif command_group[0] == 'C':
                p1 = Vector((command_group[1], command_group[2]))
                p2 = Vector((command_group[3], command_group[4]))
                p_last = Vector((command_group[5], command_group[6]))
                ret.append(['C', p1, p2, p_last])
            elif command_group[0] == 'c':
                p1 = p_last + Vector((command_group[1], command_group[2]))
                p2 = p_last + Vector((command_group[3], command_group[4]))
                p_last += Vector((command_group[5], command_group[6]))
                ret.append(['C', p1, p2, p_last])
            elif re.match(r'[Ss]', command_group[0]):
                if ret[-1][0] == 'C':
                    p1 = p_last *2 - ret[-1][2]
                else:
                    p1 = p_last
                if command_group[0] == 'S':
                    p2 = Vector((command_group[1], command_group[2]))
                    p_last = Vector((command_group[3], command_group[4]))
                else:
                    p2 = p_last + Vector((command_group[1], command_group[2]))
                    p_last += Vector((command_group[3], command_group[4]))
                ret.append(['C', p1, p2, p_last])

            # The quadratic Bezier curve commands
            elif re.match(r'[Q]', command_group[0]):
                p1 = Vector((command_group[1], command_group[2]))
                p_last = Vector((command_group[3], command_group[4]))
                ret.append([command_group[0], p1, p_last])
            elif re.match(r'[q]', command_group[0]):
                p1 = p_last + Vector((command_group[1], command_group[2]))
                p_last += Vector((command_group[3], command_group[4]))
                ret.append([command_group[0].upper(), p1, p_last])
            elif re.match(r'[Tt]', command_group[0]):
                if ret[-1][0] == 'Q':
                    p1 = p_last * 2 - ret[-1][1]
                else:
                    p1 = p_last
                if command_group[0] == 'T':
                    p_last = Vector((command_group[1], command_group[2]))
                else:
                    p_last += Vector((command_group[1], command_group[2]))
                ret.append(['Q', p1, p_last])

            #The elliptical arc curve commands
            elif re.match(r'[Aa]', command_group[0]):
                rx = abs(command_group[1])
                ry = abs(command_group[2])
                x_axis_rotation = command_group[3]
                large_arc_flag = command_group[4]
                sweep_flag = command_group[5]
                if command_group[0] == 'A':
                    p_last = Vector((command_group[6], command_group[7]))
                else:
                    p_last += Vector((command_group[6], command_group[7]))
                ret.append(['A', rx, ry, x_axis_rotation, large_arc_flag, sweep_flag, p_last])
            else:
                assert False

            if start_point_flag:
                start_point = p_last
                start_point_flag = False

        return ret



    def _tokens_to_list(self, tokens):
        last_command = ''
        command_require_num = {
                'M':2,
                'm':2,
                'Z':0,
                'z':0,
                'L':2,
                'l':2,
                'H':1,
                'h':1,
                'V':1,
                'v':1,
                'C':6,
                'c':6,
                'S':4,
                's':4,
                'Q':4,
                'q':4,
                'T':2,
                't':2,
                'A':7,
                'a':7,
                }
        ret = []
        if ('m' != tokens[0][1] != 'M'):
            warning("Strange path.")
            return []
        while tokens:
            sequence = []
            if tokens[0][0] == _D_TOKEN.COMMAND:
                command = tokens[0][1]
                del tokens[0]
            else:
                if last_command == 'M':
                    last_command = 'L'
                elif last_command == 'm':
                    last_command = 'l'
                elif last_command == 'z':
                    last_command = 'L'
                elif last_command == 'Z':
                    last_command = 'L'
                elif last_command == '':
                    warning("Strange path.")
                    return []
                command = last_command
            sequence.append(command)
            for i in xrange(command_require_num[command]):
                if tokens[0][0] != _D_TOKEN.NUMBER:
                    warning("Strange path.")
                    return []
                sequence.append(float(tokens[0][1]))
                del tokens[0]
            last_command=command

            ret.append(sequence)
        return ret

    def _tokenize(self, string):
        string = string.replace(',', ' ')
        count = 0
        ret = []
        while string:
            match_ = whitespace.match(string)
            if match_:
                string = string[match_.end():]
                if not string:
                    return None

            for tokentype, pattern in patterns:
                count+=1
                match_ = pattern.match(string)
                if match_:
                    count = 0
                    ret.append([tokentype, match_.group(0)])
                    string = string[match_.end():]
                    break
            if count >= len(patterns):
                warning("Strange path!%s" % string)
                return None
        return ret
    def __init__(self, string):
        tokens = self._tokenize(string)
        if tokens == None:
            return None
        list = self._tokens_to_list(tokens)
        if list == []:
            return None
        self.extend(self._simplify(list))
#}}}

class _PhunObject(dict):#{{{
    _default = {
            'color':[float(random()), float(random()), float(random()), 1], #FIXME
            #'entityID':1, #FIXME
            #'zDepth':1.0, #FIXME
            }
    def __str__(self):
        temp = "Scene.add%s {\n" % self.__class__.__name__
        for key, value in self.items():
            if isinstance(value, str):
                if(value != "+inf"):
                    value = '''"%s"''' % value
            elif isinstance(value, bool):
                if value:
                    value = 'true'
                else:
                    value = 'false'
            elif isinstance(value, float):
                value = "%g" % value
            temp += "    %s := %s;\n" % (key, value)
        temp += '};'
        return temp

    def __repr__(self):
        return "<%s at %x>" % (self.__class__.__name__, id(self))

    def __init__(self, attrib):
        if not MINIMUM_MODE:
            self.update(self.default)
        self.update(attrib)
#}}}

class _PhunGeometry(_PhunObject): #{{{
    default = _PhunObject._default.copy()
    default.update({
            'airFrictionMult':1.0,
            'angle':0.0,
            'attraction':0.0,
            'collideSet':1,
            'collideWater':True,
            'controllerAcc':11.0,
            'controllerInvertX':False,
            'controllerInvertY':False,
            'controllerReverseXY':False,
            'density':2.0,
            'drawBorder':True,
            'friction':0.5,
            #'geomID':1, #FIXME
            'heteroCollide':False,
            'immortal':False,
            'inertiaMultiplier':1.0,
            'killer':False,
            'materialVelocity':0.0,
            'onCollide':'(e)=>{}',
            'onHitByLaser':'(e)=>{}',
            'opaqueBorders':True,
            #'pos':[0.0, 0.0], #FIXME
            'reflectiveness':1.0,
            'refractiveIndex':1.5,
            'restitution':0.5,
            'showForceArrows':False,
            'showVelocity':False,
            'textColor':[1.0, 1.0, 1.0, 1.0],
            'textConstrained':True,
            'textFont':'Verdana',
            'textFontSize':32.0,
            'textScale':0.5,
            'texture':'',
            'textureMatrix':[1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
            })
    def get_relpos(self, point): #{{{
        return (point-self['pos']).rotate(-self['angle'])
#}}}
#}}}
class Polygon(_PhunGeometry): #{{{1
    _area = None
    _bb_x_min = None
    _bb_x_max = None
    _bb_y_min = None
    _bb_y_max = None

    default = _PhunGeometry._default.copy()
    default.update({
            })

    def _set_bb(self):#{{{
        outside_vecs = self['surfaces'][0]
        self._bb_x_min = self._bb_x_max = outside_vecs[0]
        self._bb_y_min = self._bb_y_max = outside_vecs[1]
        for point in outside_vecs:
            if point[0] < self._bb_x_min:
                self._bb_x_min = point[0]
            elif point[0] > self._bb_x_max:
                self._bb_x_max = point[0]
            if point[1] < self._bb_y_min:
                self._bb_y_min = point[1]
            elif point[1] > self._bb_y_max:
                self._bb_y_max = point[1]
#}}}
    def _get_centerpos_and_area_of_vecs(self, vecs): #{{{
        #@return 重心の座標。
        partial_area, area_sum = 0, 0;
        center_pos = Vector((0, 0));

        for i in xrange(1, len(vecs)-1):
            # ベクトル vecs[0]-vecs[i], vecs[0]-vecs[i+1]の外積を2で割った値(partial_area)を求める。
            # この値は「符号付」の三角形vec[0]、vec[i]、vec[i+1]の面積に等しい。
            partial_area = ((vecs[i]-vecs[0])[0]*(vecs[i+1]-vecs[0])[1] - (vecs[i]-vecs[0])[1]*(vecs[i+1]-vecs[0])[0])/2.

            # vec[0]、vec[i]、vec[i+1]によって作られる三角形の中心点(partial_center)を求める。
            partial_center = (vecs[0] + vecs[i] + vecs[i + 1]) / 3.;
            area_sum += partial_area;

            center_pos += partial_center * partial_area;
        if area_sum != 0:
            center_pos /= area_sum;
        return (center_pos, abs(area_sum));
        #}}}
    def _set_pos_and_area(self):#{{{
        surfaces = self['surfaces']
        each_pos = []
        each_area = []
        area = 0
        for i, vecs in enumerate(surfaces):
            temp_pos, temp_area = self._get_centerpos_and_area_of_vecs(vecs)
            if i != 0:
                temp_area = -temp_area
            each_pos.append(temp_pos)
            each_area.append(temp_area)
            area += temp_area

        pos = Vector((0, 0))
        for temp_pos, temp_area in zip(each_pos, each_area):
            pos += temp_pos * temp_area / area

        self.update({'pos':pos})
        self._area = area
#}}}
    def _current_next_pos(self, vecs): #{{{
        point_num = len(vecs)
        for i in xrange(point_num):
            yield vecs[i], vecs[(i+1)%point_num]
#}}}
    def _is_inside_of_vecs(self, vecs, point, include_on_edge): #{{{
        '''
        Return True when point is on edge or inside of polygon defined by vecs.
        @params vecs -- surfaces of the polygon
        @params point -- a survey point
        @params on_edge -- a survey point
        '''
        point_num = len(vecs)
        if point_num == 0: return False;
        if point_num == 1: return point==vecs[0]

        p_x, p_y = point[0], point[1]

        # On edge
        if include_on_edge:
            for cur, next in self._current_next_pos(vecs):
                if (
                        (
                        cur[1] == next[1] == p_y # 線分が水平で、調査点がその高さにある。
                        and
                        (cur[0] <= p_x <= next[0] or next[0] <= p_x <= cur[0]) # 線分上に調査点がある
                        )

                        or

                        (
                        cur[1] != next[1] # 線分が水平でない。
                        and
                        (cur[1] <= p_y <=next[1] or next[1] <= p_y <= cur[1]) # y座標が線分の範囲内
                        and
                        p_x == cur[0] + (next[0]-cur[0])*(p_y-cur[1])/(next[1]-cur[1]) # 線分上に調査点がある
                        )
                        ):
                    return True;

        # Inside/outside of polygon
        # 現在の点から右へ進んでいったときに何度線分と交差するかで判断
        inside = False;
        for cur, next in self._current_next_pos(vecs):
            # 現在の点が、線分に交差する時に結果を反転 (True <-> False)
            if (
                    (cur[1]<=point[1]<=next[1] or next[1]<=point[1]<=cur[1]) # y座標が線分の範囲内
                    and
                    (
                        (
                        cur[1] == next[1] # 線分が水平
                        and
                        point[0] < cur[0]
                        )
                        or
                        (
                        cur[1] != next[1] # 線分が水平でない
                        and
                        point[0] < cur[0] + (point[1]-cur[1])*(next[0]-cur[0])/(next[1]-cur[1])
                        )
                    )
                    ):
                inside = not inside
        return inside;
#}}}
    def is_include(self, point): #{{{
        pos = self['pos']
        point_rotated = pos + (pos-point).rotate(-self['angle'])
        if self._bb_x_min == None:
            self._set_bb()
        if not (self._bb_x_min <= point_rotated[0] <= self._bb_x_max and self._bb_y_min <= point_rotated[1] <= self._bb_y_max):
            return False

        for i, vecs in enumerate(self['surfaces']):
            if i == 0:
                if not self._is_inside_of_vecs(vecs, point_rotated, True): # in outside_polygon
                    return False
            else:
                if self._is_inside_of_vecs(vecs, point_rotated, False): # in hole
                    return False
        return True
#}}}

    def __init__(self, attrib):
        if not MINIMUM_MODE:
            self.update(self.default)
        self.update(attrib)
        if not 'angle' in self: self['angle'] = 0 #FIXME DRY
        self._set_pos_and_area()
#}}}1
class Box(_PhunGeometry): #{{{1
    default = _PhunGeometry._default.copy()
    default.update({
            'size':[1,0, 1,0],
            'text':'',
            'ruler':False,
            })

    def is_include(self, point): #{{{
        pos = self['pos']
        size = self['size']
        min_ = pos - size/2
        max_ = pos + size/2
        point_rotated = pos + (pos-point).rotate(-self['angle'])
        if (min_[0] < point_rotated[0] < max_[0] and min_[1] < point_rotated[1] < max_[1]):
            return True
#}}}
    def __init__(self, attrib):
        if not MINIMUM_MODE:
            self.update(self.default)
        self.update(attrib)
        if not 'angle' in self: self['angle'] = 0 #FIXME DRY
#}}}1
class Circle(_PhunGeometry): #{{{1
    def is_include(self, point): #{{{
        return self['radius'] > abs(self['pos']-point)
#}}}
    default = _PhunGeometry._default.copy()
    default.update({
            'drawCake':True,
            'radius':1.0,
            })
    def __init__(self, attrib):
        if not MINIMUM_MODE:
            self.update(self.default)
        self.update(attrib)
        if not 'angle' in self: self['angle'] = 0 #FIXME DRY
#}}}1
class Plane(_PhunGeometry): #{{{1
    default = _PhunGeometry._default.copy()
    default.update({
            })

    def is_include(self, point): #{{{
        pos = self['pos']
        point_rotated = pos + (point-pos).rotate(self['angle'])
        return point_rotated[0] < pos[0]
#}}}
    def get_relpos(self, point): #{{{
        return (point-self['pos']).rotate(self['angle'])
#}}}

    def __init__(self, attrib):
        if not MINIMUM_MODE:
            self.update(self.default)
        self.update(attrib)
        if not 'angle' in self: self['angle'] = 0 #FIXME DRY
#}}}1

class _PhunAttachment(_PhunObject):#{{{
    default = _PhunObject._default.copy()
    default.update({
            'size':0.1,
            #'opaqueBorders':True,
            })
    def _try_attatch_between2geom(self, phun_objects): #{{{2
        '''
        @param phun_objects 候補 (selfを含むリスト. 奥から手前の順)
        '''
        self_index = None
        for i, o in enumerate(phun_objects):
            if o is self:
                self_index = i

        o0, o1 = None, None
        o0_found = o1_found = False
        geom0pos = self.pop('geom0pos')
        geom1pos = self.pop('geom1pos')
        for i in xrange(self_index-1, -1, -1):
            o0 = phun_objects[i]
            if not isinstance(o0, _PhunGeometry): continue
            if o0.is_include(geom0pos):
                self['geom0'] = o0['geomID']
                self['geom0pos'] = o0.get_relpos(geom0pos)
                temp_index = i
                o0_found = True
                break
        if not o0_found:
            temp_index = self_index

        for i in xrange(temp_index-1, -1, -1):
            o1 = phun_objects[i]
            if not isinstance(o1, _PhunGeometry): continue
            if o1.is_include(geom1pos):
                self['geom1'] = o1['geomID']
                self['geom1pos'] = o1.get_relpos(geom1pos)
                o1_found = True
                break

        if o0_found and o1_found:
            return True
        elif not o0_found and not o1_found:
            return False
        else:
            if o0_found:
                self['geom1'] = 0
                self['geom1pos'] = geom1pos
                o_geom = o0
            elif o1_found:
                self['geom0'] = 0
                self['geom0pos'] = geom0pos
                o_geom = o1
            else:
                assert False
# Script below is for Algodoo's spring constant
#            if isinstance(self, Spring):
#                if isinstance(o_geom, Polygon):
#                    self['constant'] = 10 * o_geom._area
#                elif isinstance(o_geom, Box):
#                    self['constant'] = 10 * o_geom["size"][0] * o_geom["size"][1]
#                elif isinstance(o_geom, Circle):
#                    self['constant'] = 10 * 2 * pi * o_geom["radius"]**2
            return True
#}}}2
#}}}
class Spring(_PhunAttachment):#{{{
    default = _PhunAttachment._default.copy()
    default.update({
            'geom0':0,
            'geom0pos':[0.0, 0.0],
            'geom1':0,
            'geom1pos':[0.0, 0.0],
            'dampingFactor':0.1,
            'legacyMode':2,
            'length':0.0,
            'constant':0.1,
            })
    def try_attatch(self, phun_objects):
        return self._try_attatch_between2geom(phun_objects)
#}}}
class Fixjoint(_PhunAttachment):#{{{
    default = _PhunAttachment._default.copy()
    default.update({
            'geom0':-1,
            'world0pos':[0, 0],
            'geom1':0,
            'geom1ps':[0, 0],
            })
#}}}
class Hinge(_PhunAttachment):#{{{
    default = _PhunAttachment._default.copy()
    default.update({
            'geom0':-1,
            'world0pos':[0, 0],
            'geom1':-1,
            'world1pos':[0, 0],
            'ccw':False,
            'motorTorque':100.0,
            'impulseLimit':"+inf",
            'autoBrake':False,
            'motor':False,
            'motorSpeed':1.5707964,
            'distanceLimit':"+inf",
            })
    def try_attatch(self, phun_objects): #{{{2
        return self._try_attatch_between2geom(phun_objects)
#}}}2
#}}}
class Pen(_PhunAttachment): #{{{
    default = _PhunAttachment._default.copy()
    default.update({
            'fadeTime':1.5,
            'geom':1,
            'relPoint':Vector([0, 0]),
            'rotation':0
            })

    def try_attatch(self, phun_objects):
        return True #FIXME
#}}}
class LaserPen(_PhunAttachment): #{{{
    default = _PhunAttachment._default.copy()
    default.update({
            'geom':0,
            'cutter':False,
            'collideSet':127,
            'collideWater':True,
            'onLaserHit':"(e)=>{}",
            'pos':Vector([0, 0]),
            'velocity':"+inf",
            'fadeDist':25.0,
            'maxRays':1000,
            'showLaserBodyAttrib':True,
            'rotation':0.0
            })
    def try_attatch(self, phun_objects):
        return True #FIXME
#}}}

# ========================================
def string_to_rgb(color_description): #{{{
    def color_value(value):
        match_=re.search(r'^(.*)%\s*$', value)
        if match_: # "100%" や " 100 % ", "0.5%"など。
            ret = float(match_.group(1))/100.
        else:
            try:
                ret = float(value)/255.
            except:
                ret = 0
                warning("ValueError for rgb color value (%s)." % value)
        return ret

    colorkeywords = { # ColorKeywords {{{
        "aliceblue"           :"#F0F8FF",
        "antiquewhite"        :"#FAEBD7",
        "aqua"                :"#00FFFF",
        "aquamarine"          :"#7FFFD4",
        "azure"               :"#F0FFFF",
        "beige"               :"#F5F5DC",
        "bisque"              :"#FFE4C4",
        "black"               :"#000000",
        "blanchedalmond"      :"#FFEBCD",
        "blue"                :"#0000FF",
        "blueviolet"          :"#8A2BE2",
        "brown"               :"#A52A2A",
        "burlywood"           :"#DEB887",
        "cadetblue"           :"#5F9EA0",
        "chartreuse"          :"#7FFF00",
        "chocolate"           :"#D2691E",
        "coral"               :"#FF7F50",
        "cornflowerblue"      :"#6495ED",
        "cornsilk"            :"#FFF8DC",
        "crimson"             :"#DC143C",
        "cyan"                :"#00FFFF",
        "darkblue"            :"#00008B",
        "darkcyan"            :"#008B8B",
        "darkgoldenrod"       :"#B8860B",
        "darkgray"            :"#A9A9A9",
        "darkgreen"           :"#006400",
        "darkgrey"            :"#A9A9A9",
        "darkkhaki"           :"#BDB76B",
        "darkmagenta"         :"#8B008B",
        "darkolivegreen"      :"#556B2F",
        "darkorange"          :"#FF8C00",
        "darkorchid"          :"#9932CC",
        "darkred"             :"#8B0000",
        "darksalmon"          :"#E9967A",
        "darkseagreen"        :"#8FBC8F",
        "darkslateblue"       :"#483D8B",
        "darkslategray"       :"#2F4F4F",
        "darkslategrey"       :"#2F4F4F",
        "darkturquoise"       :"#00CED1",
        "darkviolet"          :"#9400D3",
        "deeppink"            :"#FF1493",
        "deepskyblue"         :"#00BFFF",
        "dimgray"             :"#696969",
        "dimgrey"             :"#696969",
        "dodgerblue"          :"#1E90FF",
        "firebrick"           :"#B22222",
        "floralwhite"         :"#FFFAF0",
        "forestgreen"         :"#228B22",
        "fuchsia"             :"#FF00FF",
        "gainsboro"           :"#DCDCDC",
        "ghostwhite"          :"#F8F8FF",
        "gold"                :"#FFD700",
        "goldenrod"           :"#DAA520",
        "gray"                :"#808080",
        "grey"                :"#808080",
        "green"               :"#008000",
        "greenyellow"         :"#ADFF2F",
        "honeydew"            :"#F0FFF0",
        "hotpink"             :"#FF69B4",
        "indianred"           :"#CD5C5C",
        "indigo"              :"#4B0082",
        "ivory"               :"#FFFFF0",
        "khaki"               :"#F0E68C",
        "lavender"            :"#E6E6FA",
        "lavenderblush"       :"#FFF0F5",
        "lawngreen"           :"#7CFC00",
        "lemonchiffon"        :"#FFFACD",
        "lightblue"           :"#ADD8E6",
        "lightcoral"          :"#F08080",
        "lightcyan"           :"#E0FFFF",
        "lightgoldenrodyellow":"#FAFAD2",
        "lightgray"           :"#D3D3D3",
        "lightgreen"          :"#90EE90",
        "lightgrey"           :"#D3D3D3",
        "lightpink"           :"#FFB6C1",
        "lightsalmon"         :"#FFA07A",
        "lightseagreen"       :"#20B2AA",
        "lightskyblue"        :"#87CEFA",
        "lightslategray"      :"#778899",
        "lightslategrey"      :"#778899",
        "lightsteelblue"      :"#B0C4DE",
        "lightyellow"         :"#FFFFE0",
        "lime"                :"#00FF00",
        "limegreen"           :"#32CD32",
        "linen"               :"#FAF0E6",
        "magenta"             :"#FF00FF",
        "maroon"              :"#800000",
        "mediumaquamarine"    :"#66CDAA",
        "mediumblue"          :"#0000CD",
        "mediumorchid"        :"#BA55D3",
        "mediumpurple"        :"#9370DB",
        "mediumseagreen"      :"#3CB371",
        "mediumslateblue"     :"#7B68EE",
        "mediumspringgreen"   :"#00FA9A",
        "mediumturquoise"     :"#48D1CC",
        "mediumvioletred"     :"#C71585",
        "midnightblue"        :"#191970",
        "mintcream"           :"#F5FFFA",
        "mistyrose"           :"#FFE4E1",
        "moccasin"            :"#FFE4B5",
        "navajowhite"         :"#FFDEAD",
        "navy"                :"#000080",
        "oldlace"             :"#FDF5E6",
        "olive"               :"#808000",
        "olivedrab"           :"#6B8E23",
        "orange"              :"#FFA500",
        "orangered"           :"#FF4500",
        "orchid"              :"#DA70D6",
        "palegoldenrod"       :"#EEE8AA",
        "palegreen"           :"#98FB98",
        "paleturquoise"       :"#AFEEEE",
        "palevioletred"       :"#DB7093",
        "papayawhip"          :"#FFEFD5",
        "peachpuff"           :"#FFDAB9",
        "peru"                :"#CD853F",
        "pink"                :"#FFC0CB",
        "plum"                :"#DDA0DD",
        "powderblue"          :"#B0E0E6",
        "purple"              :"#800080",
        "red"                 :"#FF0000",
        "rosybrown"           :"#BC8F8F",
        "royalblue"           :"#4169E1",
        "saddlebrown"         :"#8B4513",
        "salmon"              :"#FA8072",
        "sandybrown"          :"#F4A460",
        "seagreen"            :"#2E8B57",
        "seashell"            :"#FFF5EE",
        "sienna"              :"#A0522D",
        "silver"              :"#C0C0C0",
        "skyblue"             :"#87CEEB",
        "slateblue"           :"#6A5ACD",
        "slategray"           :"#708090",
        "slategrey"           :"#708090",
        "snow"                :"#FFFAFA",
        "springgreen"         :"#00FF7F",
        "steelblue"           :"#4682B4",
        "tan"                 :"#D2B48C",
        "teal"                :"#008080",
        "thistle"             :"#D8BFD8",
        "tomato"              :"#FF6347",
        "turquoise"           :"#40E0D0",
        "violet"              :"#EE82EE",
        "wheat"               :"#F5DEB3",
        "white"               :"#FFFFFF",
        "whitesmoke"          :"#F5F5F5",
        "yellow"              :"#FFFF00",
        "yellowgreen"         :"#9ACD32",
    }
    # }}}

    match_=re.search(r'rgb\(([^,]*),([^,]*),([^,]*)\)', color_description)
    if match_:
        """rgb(255,255,255)形式"""
        r = color_value(match_.group(1))
        g = color_value(match_.group(2))
        b = color_value(match_.group(3))
    else:
        if re.search(r'^#[0-9a-fA-F]{6}$', color_description):
            """#FFFFFF形式"""
            pass
        elif re.search(r'^#[0-9a-fA-F]{3}$', color_description):
            """#FFF形式"""
            temp = ""
            for s in color_description:
                temp += s + s
            color_description = temp[1:]
        else:
            """文字列による色指定(例えばredなど)の場合"""
            try:
                color_description = colorkeywords[color_description]
            except:
                warning("Color name <%s> cannot be handled." % color_description)
                color_description = "#000000"

        r = int(color_description[1:3], 16) / 255.
        g = int(color_description[3:5], 16) / 255.
        b = int(color_description[5:7], 16) / 255.
    return (r,g,b)
#}}}
# ========================================
def quadratic_bezier_to_path(p1, p2, p3, nodes = NODES): #{{{
    '''
    2次ベジエ曲線上の点を返す。
    @params p1: 始点(Vectorクラス)
    @params p2: 制御点の座標(Vectorクラス)
    @params p3: 終点(Vectorクラス)
    @params nodes(int型): 値を大きくすると点の数が増えて精度が増す。
    '''
    r = []
    for i in range(0, nodes+1):
        t = float(i) / nodes
        temp  = p1 * ((1-t)**2)
        temp += p2 * (2 * t * (1-t))
        temp += p3 * (t**2)
        r.append(Vector(temp))
    return r
#}}}
def cubic_bezier_to_path(p1, p2, p3, p4, nodes = NODES): #{{{
    '''
    3次ベジエ曲線上の点を返す。
    @params p1: 始点(Vectorクラス)
    @params p2: 第1制御点の座標(Vectorクラス)
    @params p3: 第2制御点の座標(Vectorクラス)
    @params p4: 終点(Vectorクラス)
    @params nodes(int型): 値を大きくすると点の数が増えて精度が増す。
    '''
    r = []
    for i in range(0, nodes+1):
        t = float(i) / nodes
        temp  = p1 * ((1-t)**3)
        temp += p2 * (3 * t * (1-t)**2)
        temp += p3 * (3 * t**2 * (1-t))
        temp += p4 * (t**3)
        r.append(Vector(temp))
    return r
#}}}
def elliptical_arc_to_path(p1, rx, ry, x_axis_rotation, large_arc_flag, sweep_flag, p2, nodes = NODES): #{{{
    if rx == 0 or ry ==0: return [p1, p2]
    if rx < 0: rx *= -1
    if ry < 0: ry *= -1

    x_axis_rotation *= -pi/180.
    if large_arc_flag != 0:
        large_arc_flag = 1
    if sweep_flag != 0:
        sweep_flag = 1

    p_ = ((p1-p2)/2).rotate(-x_axis_rotation)
    temp = p_[0]**2/rx**2 + p_[1]**2/ry**2
    if temp > 1:
        sqrt_temp = sqrt(temp)
        rx *= sqrt_temp
        ry *= sqrt_temp
    c_ = Vector((rx*p_[1]/ry, -ry*p_[0]/rx)) * sqrt(abs((rx*ry)**2-(rx*p_[1])**2-(ry*p_[0])**2)/((rx*p_[1])**2+(ry*p_[0])**2))
    if large_arc_flag == sweep_flag: c_ *= -1

    c = c_.rotate(x_axis_rotation) + (p1+p2)/2

    temp1 = Vector(((p_[0]-c_[0])/rx, (p_[1]-c_[1])/ry))
    temp2 = Vector(((-p_[0]-c_[0])/rx, (-p_[1]-c_[1])/ry))
    theta1 = Vector((1,0)).angle_between(temp1)
    delta_theta = temp1.angle_between(temp2) % (2 * pi)
    if sweep_flag == 0 and delta_theta > 0:
        delta_theta -= 2*pi
    elif sweep_flag == 1 and delta_theta < 0:
        delta_theta += 2*pi

    r = []
    for i in range(0, nodes+1):
        theta = float(i) / nodes * delta_theta + theta1
        p = Vector((rx*cos(theta), ry*sin(theta))).rotate(x_axis_rotation) + c
        r.append(p)
    return r
#}}}

# ========================================
# Parsers
# ========================================
def parse_text(e, m, style): # {{{
    attributes = {}
#    attributes['color'] = color
#    attributes['drawBorder'] = drawBorder

#    #より適切な位置に, 適切な大きさで。ネストに注意
#    x = float(e.get("x", 0))
#    y = float(e.get("y", 0))
#    attributes["pos"] = Vector([x,y])
#    attributes["collideSet"] = 0
#    attributes["collideWater"] = False

    #return Box(attributes)
    return None
# }}}
def parse_path(e, m, style): #{{{
    '''pathをPolygonあるいはCircleに変換'''

    # 真円の場合はCircleオブジェクトをreturn {{{
    type = e.get(sodipodi + 'type')
    if(type and type == 'arc'):
        rx = e.get(sodipodi + 'rx')
        ry = e.get(sodipodi + 'ry')

        start = e.get(sodipodi + 'start')
        end = e.get(sodipodi + 'end')
        if(rx == ry and not start and not end
                and abs(m[0])==abs(m[3]) and abs(m[1])==abs(m[2])):
            # 真円であることが確定
            cx = e.get(sodipodi + 'cx', 0)
            cy = e.get(sodipodi + 'cy', 0)
            e.set('cx', cx)
            e.set('cy', cy)
            e.set('r', rx)
            return parse_circle(e, m, style)
    #}}}

    d__ = SVG_Path(e.get('d', ''))
    if d__ == []:
        return None

    surfaces = []
    p_last = Vector([0, 0])

    while d__:
        current_command_sequence = d__.pop(0)
        if current_command_sequence[0] == 'Z':
            current_pos = p_last
        elif current_command_sequence[0] == 'M':
            current_pos = current_command_sequence[1]
        else:
            assert False

        if not d__: break
        vecs = [current_pos]
        if 'M' != d__[0][0] != 'Z':
            while d__ and 'M' != d__[0][0] != 'Z':
                current_command_sequence = d__.pop(0)
                if current_command_sequence[0] == 'L':
                    vecs.append(current_command_sequence[1])
                elif current_command_sequence[0] == 'C':
                    vecs.extend(cubic_bezier_to_path(vecs[-1], *current_command_sequence[1:]))
                elif current_command_sequence[0] == 'Q':
                    vecs.extend(quadratic_bezier_to_path(vecs[-1], *current_command_sequence[1:]))
                elif current_command_sequence[0] == 'A':
                    vecs.extend(elliptical_arc_to_path(vecs[-1], *current_command_sequence[1:]))
                else:
                    assert False
            vecs = [points.en_matrix(m) for points in vecs]
            # 同じ点が連続しないように間引く
            for i in range(len(vecs)-2, -1, -1):
                if (vecs[i] == vecs[i+1]):
                    del vecs[i]
            surfaces.append(vecs)

    if(len(surfaces) != 1):
        index = 0;
        min_x_dict = {}
        # TODO 最小の点を返す関数を作成・利用
        for vecs in surfaces:
            min_x = vecs[0][0]
            for vec in vecs:
                if vec[0] < min_x:
                    min_x = vec[0]
            min_x_dict[index] = min_x
            index+=1
        sorted_min_x = sorted(min_x_dict.items(), key=lambda x:x[1])
        sorted_min_x_index = [i[0] for i in sorted_min_x]
        surfaces = [surfaces[i] for i in sorted_min_x_index]

    color, drawBorder = style.get_color_and_drawborder()
    attributes = {
            'color':color,
            'drawBorder':drawBorder,
            'surfaces':surfaces,
            }
    return Polygon(attributes)
#}}}
def parse_polygon(e, m, style): #{{{
    '''polygonをFixjointに変換'''
    color = style.get_stroke_color()

    points = e.get('points', 0)
    match_=re.search(r'([^, ]*)[, ]*([^, ]*)[, ]*([^, ]*)[, ]*([^, ]*)[, ]*$', points)
    x0, y0 = float(match_.group(1)), float(match_.group(2))
    x1, y1 = float(match_.group(3)), float(match_.group(4))

    pos_ = Vector([x0, y0]).en_matrix(m)
    pos = Vector([x1, y1]).en_matrix(m)
    size = abs(pos_-pos)*2
    attributes = {
            'size':size,
            'color':color,
            'pos':pos,
            }
    return Fixjoint(attributes)
#}}}
def parse_rect(e, m, style): #{{{
    x = float(e.get('x', 0))
    y = float(e.get('y', 0))
    width  = float(e.get('width',  0))
    height = float(e.get('height', 0))

    try: rx = float(e.get('rx'))
    except: rx = None
    try: ry = float(e.get('ry'))
    except: ry = None
    if ((rx and ry!=0) or (ry and rx!=0)):
        # 角が丸いrect → パスとして扱う
        if not rx: rx = ry
        if not ry: ry = rx
        width2 = width -2*rx
        height2 = height -2*ry
        d  = "m %lf,%lf " % ((x+rx), y);
        d += "l %lf,0 " %  width2;
        d += "a %lf,%lf 0, 0,1 %lf,%lf"  % (rx,ry,  rx, ry);
        d += "l 0,%lf " %  height2;
        d += "a %lf,%lf 0, 0,1 %lf,%lf"  % (rx,ry, -rx, ry);
        d += "l %lf,0 " % -width2;
        d += "a %lf,%lf 0, 0,1 %lf,%lf"  % (rx,ry, -rx,-ry);
        d += "l 0,%lf " % -height2;
        d += "a %lf,%lf 0, 0,1 %lf,%lf"  % (rx,ry,  rx,-ry);
        e.set('d', d)
        return parse_path(e, m, style)

    if (m[0]*m[2]!=-m[1]*m[3]):
        # transformによりrect(直方体)でなくなっている → パスとして扱う
        p0x, p0y = x, y
        p1x, p1y = x+width, y
        p2x, p2y = x+width, y+height
        p3x, p3y = x, y+height
        d = "M %lf,%lf %lf,%lf %lf,%lf %lf,%lf z" % (p0x, p0y, p1x, p1y, p2x, p2y, p3x, p3y)
        e.set('d', d)
        return parse_path(e, m, style)

    pos = Vector([x+width/2., y+height/2.]).en_matrix(m)
    size = Vector([abs(Vector([x+width, y]).en_matrix(m) - Vector([x,y]).en_matrix(m)),
                   abs(Vector([x,y+height]).en_matrix(m) - Vector([x,y]).en_matrix(m)) ])
    angle = atan2(m[1], m[0])

    color, drawBorder = style.get_color_and_drawborder()
    attributes = {
            'color':color,
            'drawBorder':drawBorder,
            'pos':pos,
            'size':size,
            'angle':angle,
            }
    return Box(attributes)
#}}}
def parse_ellipse(e, m, style):#{{{
    cx = float(e.get('cx', 0))
    cy = float(e.get('cy', 0))
    rx = float(e.get('rx'))
    ry = float(e.get('ry'))

    if(rx == ry and abs(m[0])==abs(m[3]) and abs(m[1])==abs(m[2])):
        e.set('r', rx)
        return parse_circle(e, m, style)
    d  = "m %lf,%lf" % (cx,cy-ry);
    d += "a %lf,%lf 0, 0,1 %lf,%lf"  % (rx,ry,  rx, ry);
    d += "a %lf,%lf 0, 0,1 %lf,%lf"  % (rx,ry, -rx, ry);
    d += "a %lf,%lf 0, 0,1 %lf,%lf"  % (rx,ry, -rx,-ry);
    d += "a %lf,%lf 0, 0,1 %lf,%lf"  % (rx,ry,  rx,-ry);
    e.set('d', d)
    return parse_path(e, m, style)
#}}}
def parse_circle(e, m, style): #{{{
    cx = float(e.get('cx', 0))
    cy = float(e.get('cy', 0))
    r  = float(e.get('r'))

    if (abs(m[0])!=abs(m[3]) or abs(m[1])!=abs(m[2])):
        # transformにより円でなくなっている → ellipseとして扱う
        e.set('rx', r)
        e.set('ry', r)
        return parse_ellipse(e, m, style)

    pos = Vector([cx, cy]).en_matrix(m)
    radius = r * abs(sqrt(m[0]**2+m[1]**2))
    angle = atan2(m[1], m[0])
    color, drawBorder = style.get_color_and_drawborder()

    if("filter" in style):
        # filterの属性を持った真円はトレーサーに変換
        attributes = {
                'pos':pos,
                'size':radius*2,
                'color':color,
                }
        #for attribute in ('drawCake', 'drawBorder', 'radius' ): del attributes[attribute]
        return Pen(attributes)

    if("marker-mid" in style):
        # marker-midを持った真円はヒンジに変換
        color = style.get_stroke_color()
        attributes = {
                'geom0pos':pos,
                'geom1pos':pos,
                'color':color,
                'size':radius*2,
                }
        return Hinge(attributes)

    attributes = {
            'color':color,
            'drawBorder':drawBorder,
            'drawCake':False,
            'pos':pos,
            'radius':radius,
            'angle':angle,
            }
    return Circle(attributes)
#}}}

def parse_line(e, m, style): #{{{
    '''lineをバネまたはレーザーポインタに変換'''
    color = style.get_stroke_color()

    x1 = float(e.get('x1', 0))
    y1 = float(e.get('y1', 0))
    x2 = float(e.get('x2', 0))
    y2 = float(e.get('y2', 0))
    pos1 = Vector([x1,y1]).en_matrix(m)
    pos2 = Vector([x2,y2]).en_matrix(m)

    if 'marker-end' in style:
        direction = pos2-pos1
        attributes = {
                'color':color,
                'pos':(pos1+pos2)/2.0,
                'size':abs(pos1-pos2),
                'rotation':atan2(direction[1], direction[0]),
                }
        return LaserPen(attributes)
    else:
        attributes = {
                'color':color,
                #'strengthFactor':0.25,
                'strengthFactor':0.25,
                'geom0pos':pos1,
                'geom1pos':pos2,
                }
        return Spring(attributes)

#}}}
def parse_polyline(e, m, style): #{{{
    '''polylineを平面に変換'''
    color = style.get_stroke_color()

    points = e.get('points', 0)
    #m=re.search(r'([^, ]*)[, ]([^, ]*)[, ]*$', points)
    match_=re.search(r'([^, ]*)[, ]*([^, ]*)[, ]*([^, ]*)[, ]*([^, ]*)[, ]*$', points)

    x0, y0 = float(match_.group(1)), float(match_.group(2))
    x1, y1 = float(match_.group(3)), float(match_.group(4))
    pos = Vector([x0,y0]).en_matrix(m)
    pos_end = Vector([x1,y1]).en_matrix(m)
    direction = pos_end - pos

    attributes = {
            'drawBorder':False,
            'color':color,
            'pos':pos,
            'angle':atan2(direction[1], direction[0])
            }
    return Plane(attributes)
#}}}

# ========================================
def parse_svg(element, inherited_matrix, inherited_style, phun): #{{{
    for e in list(element):
        current_style = SVG_Style(e).inherit(inherited_style)
        current_matrix = inherited_matrix.combine(Transforms(e.get('transform')).to_matrix())

        if e.tag == (xmlns+'g'):
            parse_svg(e, current_matrix, current_style, phun)
        elif e.tag == (xmlns+'path'):
            phun.append(parse_path(e, current_matrix, current_style))
        elif e.tag == (xmlns+'rect'):
            phun.append(parse_rect(e, current_matrix, current_style))
        elif e.tag == (xmlns+'circle'):
            phun.append(parse_circle(e, current_matrix, current_style))
        elif e.tag == (xmlns+'ellipse'):
            phun.append(parse_ellipse(e, current_matrix, current_style))
        elif e.tag == (xmlns+'text'):
            pass #TODO text要素に対応
            #phun.append(parse_text(e, current_matrix, current_style))
        elif e.tag == (xmlns+'polyline'):
            phun.append(parse_polyline(e, current_matrix, current_style))
        elif e.tag == (xmlns+'line'):
            phun.append(parse_line(e, current_matrix, current_style))
        elif e.tag == (xmlns+'polygon'):
            phun.append(parse_polygon(e, current_matrix, current_style))
#}}}
# ========================================
if __name__ == '__main__': #{{{
    f = file(sys.argv[1], 'r')
    et = ElementTree(file=f)
    f.close()
    # 90で割ると、Inkscapeの1インチをPhunの1メートルに変換できる。
    scaleX, scaleY = 1/90.0, -1/90.0
    #scaleX, scaleY = 1,1
    transform = Transforms("scale(%lf,%lf)" % (scaleX, scaleY))
    style_info = SVG_Style({'fill':'black', 'fill-opacity':'1', 'stroke':'none', 'stroke-opacity':'1', 'opacity':'1'})
    phun_objects = []
    parse_svg(et.getroot(), transform.to_matrix(), style_info, phun_objects)

    for (i, o) in enumerate(phun_objects):
        if isinstance(o, _PhunGeometry):
            o['geomID'] = i+1

    #変換後のデータを出力
    print "// phun_objects scene created by %s v%s" % (sys.argv[0], VERSION)
    for o in phun_objects:
        if (isinstance(o, _PhunGeometry)): # 図形を出力
            print str(o)

    for o in phun_objects:
        if isinstance(o, _PhunAttachment): # アタッチメントを出力
            if o.try_attatch(phun_objects):
                print str(o)

    if ORIGINAL_DATA_OUTPUT:
        # 元データを出力
        print '/** Original SVG data'
        f = file(sys.argv[1], 'r')
        for line in f:
            print line,
        f.close()
        print '*/'
#}}}

# vim: expandtab shiftwidth=4 tabstop=8 softtabstop=4 fileencoding=utf-8 textwidth=99
